@@ -41,6 +41,7 @@ typedef struct {
   int min_samples;
   int max_samples;
   int sel_options;
+  int max_lock_age;
   uint32_t ref_id;
   uint32_t lock_ref_id;
   double offset;
@@ -61,14 +62,14 @@ extern void RCL_Initialise(void);
 extern void RCL_Finalise(void);
 extern int RCL_AddRefclock(RefclockParameters *params);
 extern void RCL_StartRefclocks(void);
-extern void RCL_ReportSource(RPT_SourceReport *report, struct timeval *now);
+extern void RCL_ReportSource(RPT_SourceReport *report, struct timespec *now);
 
 /* functions used by drivers */
 extern void RCL_SetDriverData(RCL_Instance instance, void *data);
 extern void *RCL_GetDriverData(RCL_Instance instance);
 extern char *RCL_GetDriverParameter(RCL_Instance instance);
 extern char *RCL_GetDriverOption(RCL_Instance instance, char *name);
-extern int RCL_AddSample(RCL_Instance instance, struct timeval *sample_time, double offset, int leap);
-extern int RCL_AddPulse(RCL_Instance instance, struct timeval *pulse_time, double second);
+extern int RCL_AddSample(RCL_Instance instance, struct timespec *sample_time, double offset, int leap);
+extern int RCL_AddPulse(RCL_Instance instance, struct timespec *pulse_time, double second);
 
 #endif
